#!/usr/bin/perl -I/usr/sausalito/perl
# $Id: /usr/sausalito/bin/em-helper.pl
#
# This is the Easy Migrate Helper Script. It's purpose is to gather all relevant
# information to migrate all or selected Vsites, Users, Emails, Cronjobs, DNS, 
# MySQL users and databases as well as important system information to another
# BlueOnyx 5209R or BlueOnyx 5210R server.
#
# For the migration this script is SCP'ed to the source of the migration and is
# executed via SSH remote command execution to report back all relevant data for 
# the migration.
#
# If you wan to poke around you can run this script yourself, but all you actually
# need to run yourself is another script, which you find here:
#
# Easy Migrate Main Script: /usr/sausalito/sbin/easy-migrate.pl
#

my $version = "005";

#
### Load required Perl modules:
#

use Getopt::Long;
use POSIX;
use Data::Dumper;
use Unix::PasswdFile;

# Debugging switch (0|1|2):
# 0 = off
# 1 = log to syslog
# 2 = log to screen
#
$DEBUG = "2";
if ($DEBUG) {
    if ($DEBUG eq "1") {
        use Sys::Syslog qw( :DEFAULT setlogsock);
    }
}

my $cceinit = '0';
my $host = hostname();
my $header_sent = '0';

# Show header:
&header;

#
### Check if we are 'root':
#
&root_check;

#
### Init:
#

my %EMDATA;

#
### Command line option handling
#

GetOptions(
  'check'       => sub { &Check() },
  'system'      => sub { &System() },
  'list'        => sub { &ListVsites() },
  'print'       => sub { &PrintListVsites() },
  'vsites'      => sub { &ExportVsites() },
  'users'       => sub { &ExportUsers() },
  'mysql'       => sub { &ExportMySQL() },
  'version'     => sub { &GetVersion() },
  'help|h|?'    => sub { help() },
) or &help();

#
### Fin:
#

print "Done!\n";

if ($cceinit == '1') {
    $cce->bye("SUCCESS");    
}
exit(0);

#
### Subs:
#

sub GetVersion {
    print "[VERSION]: " . $version . "\n\n";
}

sub BlueOnyxInit {

    use CCE;
    use Base::HomeDir qw(homedir_get_group_dir homedir_get_user_dir);
    use Sys::Hostname;
    use JSON;

    #
    ### CCE:
    #

    use CCE;
    $cce = new CCE;
    $cce->connectuds();
    $cceinit = '1';
}

sub Check {

    print "# Checking Source Platform: \n\n";

    if ((-f '/etc/build') && (-d '/usr/sausalito/codb/objects/')) {

        # Init CCE:
        if ($cceinit == '0') {
            &BlueOnyxInit();
        }

        # Get System Object:
        ($sysoid) = $cce->find('System');
        ($ok, $System) = $cce->get($sysoid);

        $FQDN = $System->{hostname} . '.' . $System->{domainname};

        # Check if compatible:
        if (($System->{productBuild} =~ /510[6-8]R/ ) || ($System->{productBuild} =~ /520[7-9]R/ ) || ($System->{productBuild} =~ /5210R/ )) {
            print "[SOURCE_OK]: Source System ($FQDN) is a " . $System->{productName} . " and is supported by Easy Migrate.\n\n";
        }
        else {
            &help("[SOURCE_FAIL]: Source System ($FQDN) is an unsupported platform! We only support imports from BlueOnyx servers.")
        }
    }
    else {
        # Not compatible!
        &help("[SOURCE_FAIL]: Source System is an unsupported platform! We only support imports from BlueOnyx servers.")
    }
}

sub System {

    print "# Getting System Data: \n\n";

    if ((-f '/etc/build') && (-d '/usr/sausalito/codb/objects/')) {

        # Init CCE:
        if ($cceinit == '0') {
            &BlueOnyxInit();
        }

        # Get System Object:
        ($sysoid) = $cce->find('System');
        ($ok, $System) = $cce->get($sysoid);

        $EMDATA = ();

        # Get the most important info from the System Object:
        $EMDATA{'System'}{'productBuild'} = $System->{productBuild};
        $EMDATA{'System'}{'productName'}  = $System->{productName};
        $EMDATA{'System'}{'hostname'}     = $System->{hostname};
        $EMDATA{'System'}{'domainname'}   = $System->{domainname};
        $EMDATA{'System'}{'FQDN'}         = $System->{hostname} . '.' . $System->{domainname};

        # Get PHP version (need this only for 5207R/5208R):
        if (-f "/home/solarspeed/php/bin/php") {
            $EMDATA{'System'}{'OSPHP'}     = `/home/solarspeed/php/bin/php -v|grep "(cli)"|awk {'print \$2'}`;
        }
        else {
            $EMDATA{'System'}{'OSPHP'}     = `/usr/bin/php -v|grep "(cli)"|awk {'print \$2'}`;
        }

        # Get List of all System NameSpaces:
        ($ok, @SystemNameSpaces) = $cce->names($sysoid);

        # Get all System NameSpaces:
        foreach my $NS (@SystemNameSpaces) {
            ($ok, $SNS) = $cce->get($sysoid, $NS);
            $EMDATA{'System'}{'NameSpaces'}{$NS} = $SNS;
        }

        # Get PHP Settings:
        ($PHPoid) = $cce->find('PHP');
        ($ok, $PHP) = $cce->get($PHPoid);
        if ($ok) {
            $EMDATA{'System'}{'PHP'} = $PHP;
        }

        # Get MYSQL Object:
        ($mysqloid) = $cce->find('MySQL');
        ($ok, $MySQL) = $cce->get($mysqloid);
        if ($ok) {
            $EMDATA{'System'}{'MySQL'} = $MySQL;
        }

        # Get ActiveMonitor Object:
        ($AMoid) = $cce->find('ActiveMonitor');
        ($ok, $ActiveMonitor) = $cce->get($AMoid);
        if ($ok) {
            $EMDATA{'System'}{'ActiveMonitor'} = $ActiveMonitor;
        }

        # Conditionally get DNS records if DNS is enabled:
        if ($EMDATA{'System'}{'NameSpaces'}{'DNS'}{'enabled'} eq '1') {

            print "DNS enabled \n";

            # Get DnsSOA:
            (@DnsSOA) = $cce->findx('DnsSOA');
            foreach my $OID (@DnsSOA) {
                ($ok, $dnsObj) = $cce->get($OID);
                $numDomFill = '0';
                if ($dnsObj->{domainname} eq "") {
                    $domainFiller = $numDomFill;
                    $numDomFill++;
                }
                else {
                    $domainFiller = $dnsObj->{domainname};

                    # Get location of DNS include file:
                    my $include_File = '/var/named/chroot/var/named/db.' . $domainFiller . '.include';
                    if (-f $include_File) {
                        $EMDATA{'DNS'}{'DnsIncludeFile'}{$domainFiller} = $include_File;
                    }
                }
                $EMDATA{'DNS'}{'DnsSOA'}{$domainFiller}->{$dnsObj->{'OID'}}   = $dnsObj;
            }

            # Get DnsRecord:
            (@DnsSOA) = $cce->findx('DnsRecord');
            foreach my $OID (@DnsSOA) {
                ($ok, $dnsObj) = $cce->get($OID);
                $EMDATA{'DNS'}{'DnsRecord'}{$dnsObj->{domainname}}->{$dnsObj->{'OID'}} = $dnsObj;
            }

            # Get DnsSlaveZone:
            (@DnsSOA) = $cce->findx('DnsSlaveZone');
            foreach my $OID (@DnsSOA) {
                ($ok, $dnsObj) = $cce->get($OID);
                $EMDATA{'DNS'}{'DnsSlaveZone'}{$dnsObj->{domain}}->{$dnsObj->{'OID'}} = $dnsObj;
            }
        }

        #
        ### Get 'WebApplications':
        #

        (@WebApplications) = $cce->findx('WebApplications');
        foreach my $OID (@WebApplications) {
            ($ok, $waObj) = $cce->get($OID);
            if ($waObj->{'appname'} ne 'roundcubemail') {
                #$EMDATA{'WebApplications'}{$waObj->{'group'}}{$waObj->{'appname'}}{$waObj->{'revision'}} = $waObj;
                $EMDATA{'WebApplications'}{$waObj->{'group'}}{$waObj->{'appname'}}{$waObj->{'sqldb'}} = $waObj;
            }
        }

        #print Dumper($EMDATA{'WebApplications'});
        print "[SYSTEM]: " . encode_json(\%EMDATA) . "\n\n";
    }
    else {
        &help("Source Platform is not a BlueOnyx!");
    }
}

sub ListVsites {

    print "# Listing Vsites: \n\n";

    if ((-f '/etc/build') && (-d '/usr/sausalito/codb/objects/')) {

        # Init CCE:
        if ($cceinit == '0') {
            &BlueOnyxInit();
        }

        # Get Vsite data:
        %VSITES = ();
        (@VsiteOIDs) = $cce->findx('Vsite');
        foreach my $OID (@VsiteOIDs) {
            ($ok, $vsiteObj) = $cce->get($OID);
            if ($ok) {
                # Makes sorting easier:
                (my $num = $vsiteObj->{name}) =~ s/site//;
                $VSITES{$num} = $vsiteObj;
            }
        }

        # Pretty print the output:
        print "┌──────────────┬───────────────────────────────────────────┬────────────────────┬─────────────────────────────────────────┐\n";
        print "│ Vsite Group  │       Vsite Domain Name                   │  IPv4 IP Address   │             IPv6 IP Address             │\n";
        print "├──────────────┼───────────────────────────────────────────┼────────────────────┼─────────────────────────────────────────┤\n";

        foreach $key (sort { $a <=> $b } keys(%VSITES)) {
            $line = sprintf "%-11s │  %-40s │  %-16s  │  %-36s   ", $VSITES{$key}{'name'}, $VSITES{$key}{'fqdn'}, $VSITES{$key}{'ipaddr'}, $VSITES{$key}{'ipaddrIPv6'};
            print "│  " . $line . "│\n";
        }
        print "└──────────────┴───────────────────────────────────────────┴────────────────────┴─────────────────────────────────────────┘\n\n";

    }
    else {
        &help("Source Platform is not a BlueOnyx!");
    }
}

sub PrintListVsites {

    print "# Listing Vsites: \n\n";

    if ((-f '/etc/build') && (-d '/usr/sausalito/codb/objects/')) {

        # Init CCE:
        if ($cceinit == '0') {
            &BlueOnyxInit();
        }

        # Get Vsite data:
        %VSITES = ();
        (@VsiteOIDs) = $cce->findx('Vsite');
        foreach my $OID (@VsiteOIDs) {
            ($ok, $vsiteObj) = $cce->get($OID);
            if ($ok) {
                # Makes sorting easier:
                (my $num = $vsiteObj->{name}) =~ s/site//;
                $VSITES{$num}{$vsiteObj->{name}}{'Vsite'} = $vsiteObj;
            }
        }

        # Sort - not that it matters:
        my $VSITEDATA = ();
        foreach $key (sort { $a <=> $b } keys(%VSITES)) {
            $VSITEDATA{$key} = $VSITES{$key};
        }

        # Return json:
        print "[VSITES_SHORT]: " . encode_json(\%VSITEDATA) . "\n\n";

    }
    else {
        &help("Source Platform is not a BlueOnyx!");
    }
}

sub ExportVsites {

    print "# Exporting Vsites: \n\n";

    if ((-f '/etc/build') && (-d '/usr/sausalito/codb/objects/')) {

        # Init CCE:
        if ($cceinit == '0') {
            &BlueOnyxInit();
        }

        # Get Vsite data:
        %VSITES = ();
        (@VsiteOIDs) = $cce->findx('Vsite');
        foreach my $OID (@VsiteOIDs) {

            ($ok, $vsiteObj) = $cce->get($OID);
            if ($ok) {
                # Makes sorting easier:
                (my $num = $vsiteObj->{name}) =~ s/site//;
                $VSITES{$num}{$vsiteObj->{name}}{'Vsite'} = $vsiteObj;

                # Create an Index that associates $vsiteObj->{fqdn} with $vsiteObj->{name}:
                $VSITES{'INDEX'}{$vsiteObj->{fqdn}} = $vsiteObj->{name};
                $VSITES{'NAME_INDEX'}{$vsiteObj->{name}} = $vsiteObj->{fqdn};

                # Get List of all Vsite NameSpaces:
                ($ok, @VsiteNameSpaces) = $cce->names($OID);

                # Get all Vsite NameSpaces:
                foreach my $NS (@VsiteNameSpaces) {
                    ($ok, $SNS) = $cce->get($OID, $NS);
                    $VSITES{$num}{$vsiteObj->{name}}{'NameSpaces'}{$NS}   = $SNS;

                    # Do a WebApp Backup:
                    if (($SNS->{'appname'}) && ($SNS->{'isactive'} eq "1")) {

                        # Update the main objext of 'WebApplications' to trigger an SQL-Dump of the WebApp DB:
                        (@WA_OIDS) = $cce->find('WebApplications', { 'appname' => $SNS->{'appname'}, 'group' => $vsiteObj->{name}, 'sqldb' => $SNS->{'sqldb'} });
                        if (scalar(@WA_OIDS)) {
                            ($ok) = $cce->set($WA_OIDS[0], '', { 'doBackupDB' => time() });
                        }
                        if (!$ok) {
                            print "[WARN]: Unable to make SQL backup of WebApp '$NS' for Vsite $vsiteObj->{'fqdn'}.\n";
                        }
                        else {
                            print "[INFO]: Created SQL backup of WebApp '$NS' for Vsite $vsiteObj->{'fqdn'}.\n";
                        }
                    }

                    # Conditionally export Vsite MySQL databases:
                    if (($NS eq "MYSQL_Vsite") && ($SNS->{'enabled'} eq "1") && ($SNS->{'enabled'} eq "1") && ($SNS->{'username'} ne "") && ($SNS->{'pass'} ne "") && ($SNS->{'DB'} ne "")) {
                        if ($System->{productBuild} =~ /5210R/ ) {
                            $sql_exportDir = $vsiteObj->{'basedir'} . '/wwwroot/sql/';
                        }
                        else {
                            $sql_exportDir = $vsiteObj->{'basedir'} . '/sql/';
                        }
                        $vsite_group = $vsiteObj->{'name'};
                        if (! -d $sql_exportDir) {
                            system("mkdir -p $sql_exportDir");
                            system("chmod 0755 $sql_exportDir");
                            system("chown nobody:$vsite_group $sql_exportDir");
                        }

                        @DBmulti = $cce->scalar_to_array($SNS->{'DBmulti'});

                        # Get MYSQL Object:
                        ($mysqloid) = $cce->find('MySQL');
                        ($ok, $MySQL) = $cce->get($mysqloid);
                        if ($ok) {
                            if (($MySQL->{'sql_rootpassword'} ne "") && ($MySQL->{'sql_root'} ne "")) {
                                #print "/usr/bin/mysqldump --opt --routines --extended-insert=false -c -Q --user=" . $MySQL->{'sql_root'} . " --password=" . $MySQL->{'sql_rootpassword'} . " " . $SNS->{'DB'} . " into " . $sql_exportDir . $SNS->{'DB'} . ".sql\n";
                                system("/usr/bin/mysqldump --opt --routines --extended-insert=false -c -Q --user=" . $MySQL->{'sql_root'} . " --password=" . $MySQL->{'sql_rootpassword'} . " " . $SNS->{'DB'} . " > " . $sql_exportDir . $SNS->{'DB'} . ".sql");
                                print "[INFO]: Created SQL backup of '$NS' database '" . $SNS->{'DB'} . "' for Vsite $vsiteObj->{'fqdn'}.\n";

                                if ((scalar(@DBmulti)) > "0") {
                                    foreach my $extra_db (@DBmulti) {
                                        #print "/usr/bin/mysqldump --opt --routines --extended-insert=false -c -Q --user=" . $MySQL->{'sql_root'} . " --password=" . $MySQL->{'sql_rootpassword'} . " " . $extra_db . " into " . $sql_exportDir . $extra_db . ".sql\n";
                                        system("/usr/bin/mysqldump --opt --routines --extended-insert=false -c -Q --user=" . $MySQL->{'sql_root'} . " --password=" . $MySQL->{'sql_rootpassword'} . " " . $extra_db . " > " . $sql_exportDir . $extra_db . ".sql");
                                        print "[INFO]: Created SQL backup of '$NS' database '" . $extra_db . "' for Vsite $vsiteObj->{'fqdn'}.\n";
                                    }
                                }
                            }
                        }
                        system("chown -R nobody:$vsite_group $sql_exportDir");
                    }
                }
            }
        }

        # Sort - not that it matters:
        my $VSITEDATA = ();
        foreach $key (sort { $a <=> $b } keys(%VSITES)) {
            $VSITEDATA{$key} = $VSITES{$key};
        }

        # Return json:
        print "[VSITES]: " . encode_json(\%VSITEDATA) . "\n\n";
    }
    else {
        &help("Source Platform is not a BlueOnyx!");
    }
}

sub ExportUsers {

    print "# Exporting Users: \n\n";

    if ((-f '/etc/build') && (-d '/usr/sausalito/codb/objects/')) {

        #
        ### Get User System passwords:
        #

        %PASSWD = ();
        if (-f '/etc/shadow') {
            $pw_file = '/etc/shadow';

            # Open password file:
            open(PASSWD, $pw_file);
            while (<PASSWD>) {
                chomp;
                ($login, $passwd, $uid, $gid, $gcos, $home, $shell) = split(/:/);
                if (($passwd ne '!!') && ($passwd ne '*')) {
                    $PASSWD->{$login} = $passwd;
                }
            }
        }

        # Init CCE:
        if ($cceinit == '0') {
            &BlueOnyxInit();
        }

        # Get User data:
        %USERS = ();
        (@VsiteOIDs) = $cce->findx('User');
        foreach my $OID (@VsiteOIDs) {
            ($ok, $vsiteObj) = $cce->get($OID);
            if ($ok) {
                if ($vsiteObj->{'site'} eq '') {
                    $vsiteObj->{'site'} = 'users';
                }

                # Adjust Password and store:
                $vsiteObj->{'md5_password'} = $PASSWD->{$vsiteObj->{'name'}};
                $USERS{$vsiteObj->{'site'}}{$vsiteObj->{'name'}}{'User'} = $vsiteObj;

                # Get List of all User NameSpaces:
                ($ok, @VsiteNameSpaces) = $cce->names($OID);

                # Get all User NameSpaces:
                foreach my $NS (@VsiteNameSpaces) {
                    ($ok, $SNS) = $cce->get($OID, $NS);
                    $USERS{$vsiteObj->{'site'}}{$vsiteObj->{'name'}}{'NameSpaces'}{$NS}   = $SNS;
                }

                # Parse /etc/passwd to find out where the User has his home directory:
                $pw = new Unix::PasswdFile "/etc/passwd";
                $USERS{$vsiteObj->{'site'}}{$vsiteObj->{'name'}}{'HomeDirectory'} = $pw->home($vsiteObj->{'name'});
                undef $pw;
            }

            # Get User-Cronjobs:
            my $usercron = '/var/spool/cron/' . $vsiteObj->{'name'};
            if (-f $usercron) {
                $USERS{$vsiteObj->{'site'}}{$vsiteObj->{'name'}}{'UserCron'}{$usercron}   = $usercron;
            }
        }

        # Sort - not that it matters:
        my $USERDATA = ();
        foreach $key (sort { $a <=> $b } keys(%USERS)) {
            $USERDATA{$key} = $USERS{$key};
        }

        # Return json:
        print "[USERS]: " . encode_json(\%USERDATA) . "\n\n";

    }
    else {
        &help("Source Platform is not a BlueOnyx!");
    }
}

sub ExportMySQL {

    # Directory setup:
    system("mkdir -p /home/.easy-migrate/");
    system("chmod 0700 /home/.easy-migrate/");

    # Init CCE:
    if ($cceinit == '0') {
        &BlueOnyxInit();
    }

    ($mysqloid) = $cce->find('MySQL');
    ($ok, $MySQL) = $cce->get($mysqloid);
    if ($ok) {
        my $mysqlroot = $MySQL->{'sql_root'};
        my $mysqlpass = $MySQL->{'sql_rootpassword'};
        system("mysql -u $mysqlroot -p$mysqlpass --skip-column-names -A -e\"SELECT CONCAT('SHOW GRANTS FOR ''',user,'''\@''',host,''';') FROM mysql.user WHERE user NOT IN ('root','avspam')\" | mysql -u $mysqlroot -p$mysqlpass --skip-column-names -A | sed 's/\$/;/g' > /home/.easy-migrate/mysqlusergrants.sql");
        my $dblist = `mysql -u $mysqlroot -p$mysqlpass -ANe\"SELECT schema_name FROM information_schema.schemata WHERE schema_name NOT IN ('mysql','test','information_schema','performance_schema','avspam6')\"`;
        $dblist = join " ", split "\n", $dblist;
        system("mysqldump -u $mysqlroot -p$mysqlpass --routines --triggers --single-transaction --databases $dblist > /home/.easy-migrate/mysql-all-dbs.sql");
        system("chmod 0600 -R /home/.easy-migrate/*.sql");
    }
    else {
        &help("Unable to get MySQL access details out of CCEd.");
    }
    print "[MYSQL_OK]: Data exported.\n\n";
}

sub header {
    print "┌──────────────────────────────────────────────────────────────┐\n";
    print "│ Easy Migrate Helper: Easily Migrate between BlueOnyx Servers │\n";
    print "└──────────────────────────────────────────────────────────────┘\n\n";
    $header_sent++;
}

sub help {
    $error = shift || "";
    if ($header_sent eq '0') {
        &header;
    }
    if ($error) {
        print "ERROR: $error\n\n";
    }
    print "usage:   em-helper.pl [OPTION]\n";
    print "\n";
    print "         --check      Test if source platform is suitable as migration source.\n";
    print "         --list       Human readable list of Vsites present on source server.\n";
    print "         --print      Short JSON Hash of Vsites present on source server.\n";
    print "         --system     Export full JSON Hash of System relevant data.\n";
    print "         --vsites     Export full JSON Hash of Vsites present on source server.\n";
    print "         --users      Export full JSON Hash of Users present on source server.\n";
    print "         --mysql      Export MySQL data to /home/.easy-migrate/\n";
    print "         --version    Show Version of this script.\n";
    print "         -h|--help    This help text\n\n";
    if ($cceinit == '1') {
        $cce->bye("SUCCESS");    
    }
    exit(1);
}

sub root_check {
    my $id = `id -u`;
    chomp($id);
    if ($id ne "0") {
        &help("$0 must be run by user 'root'!");
    }
}

sub debug_msg {
    if ($DEBUG eq "1") {
        $msg = shift;
        setlogsock('unix');
        openlog($0,'','user');
        syslog('info', "$ARGV[0]: $msg");
        closelog;
    }
    if ($DEBUG eq "2") {
        my $msg = shift;
        print $msg;
    }
}

# 
# Copyright (c) 2008-2019 Michael Stauber, SOLARSPEED.NET
# Copyright (c) 2008-2019 Team BlueOnyx, BLUEONYX.IT
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#     notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in 
#     the documentation and/or other materials provided with the 
#     distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#     contributors may be used to endorse or promote products derived 
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 